"""
CLI App to download mySociety datasets
"""
import json
import urllib.request
from typing import Any, List

import rich
import rich_click as click
from rich.panel import Panel

from .dataset import DataRepo, get_dataset_options, get_dataset_url, get_public_datasets


class PanelPrint:
    """
    Helper for printing list of items in a panel
    """

    def __init__(self, expand: bool = False, **kwargs: Any) -> None:
        self.items: List[str] = []
        self.panel_properties = kwargs
        self.panel_properties["expand"] = expand

    def print(self, item: str) -> None:
        """
        Add item to panel
        """
        self.items.append(item)

    def display(self) -> None:
        """
        Display panel
        """
        panel = Panel("\n".join(self.items), **self.panel_properties)
        rich.print(panel)


@click.group()  # type: ignore
def cli():
    """
    cli manager for mysoc dataset
    """


@cli.command("list")
@click.option("--repo", help="Data repo name", default="")
@click.option("--package", help="Package name", default="")
@click.option("--version", help="Version name", default="")
@click.option("--json", "use_json", is_flag=True, help="Output as JSON")
def do_list(
    repo: str = "", package: str = "", version: str = "", use_json: bool = False
):
    """
    Depending which options have a non blank value, display a list of avaliable repos,
    packages, versions, or files.
    """
    # can't have a set version and not a set package and repo
    if version and not (package and repo):
        raise click.ClickException("Can't set version without package and repo")
    # can't have a set package and not a set repo
    if package and not repo:
        raise click.ClickException("Can't set package without repo")

    if repo == "":
        title = "Public data repos"
        items = get_public_datasets()
    elif package == "":
        title = f"Packages in {repo}"
        items = DataRepo(repo).list_packages()
    elif version == "":
        title = f"Versions in {repo},{package}"
        items = DataRepo(repo).get_package(package).list_versions()
    else:
        title = f"Files in {repo},{package},{version}"
        items = DataRepo(repo).get_package(package).get_version(version).get_filenames()

    # create a rich
    if use_json:
        rich.print(json.dumps(items))
    else:
        p = PanelPrint(
            padding=1,
            style="blue",
            title=title,
            subtitle="See more options with `--help`",
        )
        for item in items:
            p.print(f"[green]{item}[/green]")
        p.display()


@cli.command("survey")
@click.option("--repo", help="Data repo name")
@click.option("--package", help="Package name")
@click.option("--version", help="Version name", default="latest")
@click.option("--file", "filename", help="file name")
def survey(
    repo: str,
    package: str,
    version: str,
    filename: str,
):
    """
    Retrieve and print the survey url of a file
    """
    if not repo:
        raise click.ClickException("Must provide repo")
    if not package:
        raise click.ClickException("Must provide package")
    if not version:
        raise click.ClickException("Must provide version")
    if not filename:
        raise click.ClickException("Must provide file")

    data_package_url = get_dataset_options(repo, package, version, filename)
    rich.print(data_package_url.survey_link)


@cli.command("url")
@click.option("--repo", help="Data repo name")
@click.option("--package", help="Package name")
@click.option("--version", help="Version name", default="latest")
@click.option("--file", "filename", help="file name")
@click.option("--quiet", "quiet", is_flag=True, help="Only print the url")
def get_url(
    repo: str,
    package: str,
    version: str,
    filename: str,
    quiet: bool,
):
    """
    Retrieve and print the url of a file
    """
    if not repo:
        raise click.ClickException("Must provide repo")
    if not package:
        raise click.ClickException("Must provide package")
    if not version:
        raise click.ClickException("Must provide version")
    if not filename:
        raise click.ClickException("Must provide file")

    data_package_url = get_dataset_url(
        repo,
        package,
        version,
        filename,
        done_survey=quiet,
        ignore_version_warning=quiet,
    )
    rich.print(data_package_url)


@cli.command()
@click.option("--repo", help="Data repo name")
@click.option("--package", help="Package name")
@click.option("--version", help="Version name", default="latest")
@click.option("--file", "filename", help="file name")
@click.option(
    "--dest-dir", "dest_path", help="Destination directory (optional)", default=""
)
@click.option("--done-survey", help="Turns off nag prompt", is_flag=True)
def download(
    repo: str,
    package: str,
    version: str,
    filename: str,
    dest_path: str,
    done_survey: bool = False,
):
    """
    Retrieve and print the url of a file
    """
    if not repo:
        raise click.ClickException("Must provide repo")
    if not package:
        raise click.ClickException("Must provide package")
    if not version:
        raise click.ClickException("Must provide version")
    if not filename:
        raise click.ClickException("Must provide file")

    data_package_url = get_dataset_url(
        repo, package, version, filename, done_survey=done_survey
    )

    # extract file name from url
    file_name = data_package_url.split("/")[-1]

    if dest_path:
        file_name = dest_path + "/" + file_name
    else:
        file_name = "./" + file_name

    # download file
    urllib.request.urlretrieve(data_package_url, file_name)
    rich.print(f"[green]Downloaded file[/green] to {file_name}")


@cli.command("detail")
@click.option("--repo", help="Data repo name")
@click.option("--package", help="Package name")
@click.option("--version", help="Version name", default="latest")
def detail(
    repo: str,
    package: str,
    version: str,
):
    """
    Retrieve and print the frictionless datapackage.json file
    """
    if not repo:
        raise click.ClickException("Must provide repo")
    if not package:
        raise click.ClickException("Must provide package")
    if not version:
        raise click.ClickException("Must provide version")

    data_package_url = get_dataset_url(
        repo, package, version, "datapackage.json", done_survey=True
    )
    # open url and load package into json
    with urllib.request.urlopen(data_package_url) as response:
        data_package = json.load(response)
    rich.print(data_package)


def main():
    """
    Main function to run
    """
    cli()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['mysoc_dataset']

package_data = \
{'': ['*']}

install_requires = \
['rich-click>=1.5.2,<2.0.0', 'rich>=12.5.1,<13.0.0']

extras_require = \
{':python_version < "3.8"': ['pandas==1.3.5'],
 ':python_version >= "3.8"': ['pandas>=1.3.5']}

entry_points = \
{'console_scripts': ['mysoc-dataset = mysoc_dataset.__main__:main']}

setup_kwargs = {
    'name': 'mysoc-dataset',
    'version': '0.2.2',
    'description': 'Tool to download mySociety datasets',
    'long_description': '# mysoc-dataset\n\nA python package and CLI to download mySociety datasets.\n\nThe best example of one of these datasets is: https://mysociety.github.io/uk_local_authority_names_and_codes/\n\nThese datasets are versioned [frictionless datapackages](https://frictionlessdata.io/). A repository may contain multiple data packages, each data package may contain multiple versions. A version of a datapackage will contain multiple resources (.csvs) and some composite files made up of resources (.xlsx or .sqlite). \n\n\n## Installation\n\nIt can be installed with:\n\n```\npython -m pip install mysoc-dataset\n```\n\nor if using poetry:\n\n```\npoetry add mysoc-dataset\n```\n\n## Usage\n\n### As a package\n\nThe package has two basic functions (with helpful error messages) to access the url or a dataframe\nof the resource.\n\n```python\nfrom mysoc_dataset import get_dataset_url, get_dataset_df\n\nurl = get_dataset_url(\n    repo_name="uk_local_authority_names_and_codes",\n    package_name="uk_la_future",\n    version_name="latest",\n    file_name="uk_local_authorities_future.csv",\n)\n\n# get a pandas dataframe\ndf = get_dataset_df(\n    repo_name="uk_local_authority_names_and_codes",\n    package_name="uk_la_future",\n    version_name="latest",\n    file_name="uk_local_authorities_future.csv",\n)\n\n\n```\n\n### As a CLI\n\nThe CLI can be used to explore avaliable data using the `list` command, get the [frictionless datapackage](https://frictionlessdata.io/) that describes the repo using `detail` fetch the url with the `url` command or download the file using `download`.\n\nThis can be used to source files or pipe the URLs into other functions without writing python scripts. \n\nThe CLI can either be run as `python -m mysoc_dataset` or `mysoc-dataset`. \n\nFor instance, the following will print the `datapackage.json` that describes the underlying contents. \n\n`mysoc-dataset detail --repo uk_local_authority_names_and_codes --version latest --package uk_la_future`\n\nAnd the following will get the URL of the resource, pegged to the `1` major version:\n\n`mysoc-dataset url --repo uk_local_authority_names_and_codes --version 1 --package uk_la_future --file uk_local_authorities_future.csv`\n\nIf the dataset has had a major change, a warning will indicate this is no longer the latest version - while not introducing breaking changes to headers without the script being changed.\n\nUse `mysoc-dataset --help` for more instructions. \n\nIf using the CLI for a dataset, please fill out a survey of what you are using it for to help us explain the value of the data to funders. You can get a URL to the survey page using the \'survey\' command. \n\n`mysoc-dataset survey --repo uk_local_authority_names_and_codes --version latest --package uk_la_future --file uk_local_authorities_future.csv`\n\n# Maintenance\n\nIf the repo has a valid PYPI_TOKEN secret, and if the poetry version is bumped and all tests pass - the GitHub Action will automatically publish on push to the main branch.',
    'author': 'Alex Parsons',
    'author_email': 'alex.parsons@mysociety.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/mysociety/mysoc-dataset',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



import functools
import json
import urllib.error
import urllib.request
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any, Callable, Dict, List

import pandas as pd
import rich

# URL to consult for public repos and default settings
PUBLIC_URL_LIST = "https://data.mysociety.org/datarepos.json"

# Default domain to fall back on if lookup to above fails
DEFAULT_DATAREPO_DOMAIN = "https://mysociety.github.io"


def timed_cache(**timedelta_kwargs: Any):
    """
    Cache with timedelta timeout
    """

    def _wrapper(f: Callable[..., Any]):
        update_delta = timedelta(**timedelta_kwargs)
        next_update = datetime.utcnow() + update_delta
        # Apply @lru_cache to f with no cache size limit
        f = functools.lru_cache(None)(f)

        @functools.wraps(f)
        def _wrapped(*args: Any, **kwargs: Any):
            nonlocal next_update
            now = datetime.utcnow()
            if now >= next_update:
                f.cache_clear()
                next_update = now + update_delta
            return f(*args, **kwargs)

        return _wrapped

    return _wrapper


class RepoNotFound(Exception):
    """
    Error to raise if REPO_URL/data.json is not found
    """


class PackageNotFound(Exception):
    """
    Error to raise if a package has been given but isn't avaliable in that repo
    """


class VersionNotFound(Exception):
    """
    Error to raise if a specified version of a package is not avaliable
    """


class FileNotFound(Exception):
    """
    Error to raise if a specified file in a verison is not avaliable
    """


def valid_url(url: str) -> bool:
    """
    check if url is a URL format
    """
    if "https://" in url or "http://" in url:
        return True
    return False


@timed_cache(minutes=1)
def fetch_data_repo(url: str) -> Dict[str, Any]:
    """
    retrieve the data.json file from the data repo
    """
    try:
        with urllib.request.urlopen(url) as url_conn:
            data = json.loads(url_conn.read().decode())
            return data
    except urllib.error.HTTPError as exc:
        raise RepoNotFound(
            f"Data repo not found: {url}. Available repos: {get_public_datasets()}"
        ) from exc


def uncached_get_datarepos_json() -> Dict[str, Any]:
    """
    Retrieve the public URL list - but fail softly and return a {} if nothing found
    """
    try:
        with urllib.request.urlopen(PUBLIC_URL_LIST) as url:
            data = json.loads(url.read().decode())
            return data
    except urllib.error.HTTPError:
        rich.print(
            "[red]Could not access the datarepos.json file. Using default configurations[/red]"
        )
        return {}


# version that is cached for a minute
get_datarepos_json = timed_cache(minutes=1)(uncached_get_datarepos_json)


def get_default_domain() -> str:
    """
    Get the default domain for the data repos
    """
    return get_datarepos_json().get("default_domain", DEFAULT_DATAREPO_DOMAIN)


def get_public_datasets() -> List[str]:
    """
    Fetch the public url list, which is a list of URLs under a "datarepos" key.
    and return this list
    """
    return get_datarepos_json().get("datarepos", [])


@dataclass
class Download:
    """
    A downloadable resource
    """

    slug: str
    url: str
    survey_link: str

    @classmethod
    def load(cls, slug: str, data: Dict[str, Any]) -> "Download":
        """
        Load a download from a dictionary
        """
        return Download(
            slug=slug,
            url=data["url"],
            survey_link=data["survey_link"],
        )


@dataclass
class PackageVersion:
    """
    A collection of downloadable resources associated with a version of a package
    """

    version: str  # short version (may be '1', or 'latest')
    full_version: str  # semvar version
    downloads: Dict[str, Download]

    @classmethod
    def load(cls, slug: str, data: Dict[str, Any]) -> "PackageVersion":
        """
        Load a package version from a dictionary
        """
        return PackageVersion(
            version=slug,
            full_version=data["full_version"],
            downloads={
                slug: Download.load(slug, download)
                for slug, download in data["files"].items()
            },
        )

    def get_filenames(self) -> List[str]:
        """
        Get all downloads
        """
        urls = [x.url for x in self.downloads.values()]
        # extract final filename from url
        return [x.split("/")[-1] for x in urls]

    def get_file(self, slug: str) -> Download:
        """
        Get a download by slug
        """
        try:
            return self.downloads[slug]
        except KeyError as exc:
            raise FileNotFound(
                f"File not found: {slug}. Available files: {list(self.downloads.keys())}"
            ) from exc


@dataclass
class Package:
    """
    A collection of versions of the same package
    """

    slug: str
    latest_version: str
    versions: Dict[str, PackageVersion]

    @classmethod
    def from_data(cls, slug: str, data: Dict[str, Any]):
        """
        Load a package from a dictionary
        """
        return Package(
            slug=slug,
            latest_version=data["latest_version"],
            versions={
                slug: PackageVersion.load(slug, version)
                for slug, version in data["versions"].items()
            },
        )

    def list_versions(self) -> List[str]:
        """
        List all versions of a package
        """
        return list(self.versions.keys())

    def get_version(
        self, version: str, ignore_version_warning: bool = False
    ) -> PackageVersion:
        """
        Get a specific version of a package
        """
        if version not in self.versions:
            raise VersionNotFound(
                f"Version {version} not found. Available versions are {self.list_versions()}"
            )
        version_obj = self.versions[version]

        # raise warning if version is not a full 'x.x.x' semver and version is not self.latest_version
        if version_obj.full_version != self.latest_version:
            if not ignore_version_warning:
                rich.print(
                    f"[yellow]Version {version} of {self.slug} is not the latest version. Latest version is {self.latest_version}[/yellow]"
                )

        return version_obj


class DataRepo:
    """
    Management for the DataRepo level

    Given either a full url, or a data repo name
    Fetch and unpack the data.json to understand resoruces
    """

    def __init__(self, data_repo_ref: str):
        if valid_url(data_repo_ref):
            self.url = data_repo_ref
        else:
            self.url = f"{get_default_domain()}/{data_repo_ref}"

        self.data = fetch_data_repo(f"{self.url}/data.json")
        self.packages: Dict[str, Package] = {}
        for slug, package in self.data.items():
            self.packages[slug] = Package.from_data(slug, package)

    def list_packages(self) -> List[str]:
        """
        List all packages in the data repo
        """
        return list(self.packages.keys())

    def get_package(self, slug: str) -> Package:
        """
        Get a package by slug. Raise PackageNotFound if not found.
        """
        try:
            return self.packages[slug]
        except KeyError as exc:
            raise PackageNotFound(
                f"Package {slug} not found. Available packages: {self.list_packages()}"
            ) from exc


def get_dataset_options(
    repo_name: str,
    package_name: str,
    version_name: str,
    file_name: str,
    ignore_version_warning: bool = False,
) -> Download:
    """
    Get a URL for a specific file in a specific version of a package in a specific data repo
    """
    repo = DataRepo(repo_name)
    package = repo.get_package(package_name)
    version = package.get_version(version_name, ignore_version_warning)
    file = version.get_file(file_name)
    return file


def get_dataset_url(
    repo_name: str,
    package_name: str,
    version_name: str,
    file_name: str,
    ignore_version_warning: bool = False,
    done_survey: bool = False,
) -> str:
    """
    Get a URL for a specific file in a specific version of a package in a specific data repo.

    Args:
        repo_name (str): The name of the data repo either the subfolder of the mysociety default domain or a full URL
        package_name (str): The name of the datapackage in the repo
        version_name (str): The name of the version - may be a full semvar version, a major or minor version, or 'latest'
        file_name (str): The name of the file to download that is part of the package (usually a .csv)
        ignore_version_warning (bool, optional): Ignore the warning if the version is not the latest. Defaults to False.
        done_survey (bool, optional): Have you completed the survey? Defaults to False. Nagging message displayed if not.

    Returns:
        str: The URL to download the file from


    """
    dataset_obj = get_dataset_options(
        repo_name, package_name, version_name, file_name, ignore_version_warning
    )
    if done_survey is False:
        rich.print(
            f"If you find [blue]{package_name}[/blue] helpful, can you tell us how using this survey? {dataset_obj.survey_link}. This message can be removed by setting the `done_survey` option to True."
        )
    return dataset_obj.url


def get_dataset_df(
    repo_name: str,
    package_name: str,
    version_name: str,
    file_name: str,
    ignore_version_warning: bool = False,
    done_survey: bool = False,
) -> pd.DataFrame:
    """
    Get a pandas dataframe for a specific file in a specific version of a package in a specific data repo.
    Get a URL for a specific file in a specific version of a package in a specific data repo.

    Args:
        repo_name (str): The name of the data repo either the subfolder of the mysociety default domain or a full URL
        package_name (str): The name of the datapackage in the repo
        version_name (str): The name of the version - may be a full semvar version, a major or minor version, or 'latest'
        file_name (str): The name of the file to download that is part of the package (usually a .csv)
        ignore_version_warning (bool, optional): Ignore the warning if the version is not the latest. Defaults to False.
        done_survey (bool, optional): Have you completed the survey? Defaults to False. Nagging message displayed if not.

    Returns:
        pd.DataFrame: The dataframe of the file
    """
    dataset_url = get_dataset_url(
        repo_name,
        package_name,
        version_name,
        file_name,
        ignore_version_warning,
        done_survey,
    )
    return pd.read_csv(dataset_url)  # type:ignore



"""
Tool to download mySociety datasets
"""
__version__ = "0.2.2"
from .dataset import get_dataset_df, get_dataset_url, get_public_datasets

__all__ = ["get_dataset_url", "get_public_datasets", "get_dataset_df"]
